use crate::{datastore::State, grpc::challenge::Keys};
use serde::{Deserialize, Serialize};
use serde_with::base64::Base64;

#[serde_with::serde_as]
#[derive(Serialize, Deserialize)]
pub struct KeysFile {
    random: String,
    #[serde_as(as = "Base64")]
    keypair: Vec<u8>,
    token: String,
}

impl From<Keys> for KeysFile {
    fn from(keys: Keys) -> Self {
        use rand::{distributions::Alphanumeric, Rng};
        let random: String =
            rand::thread_rng().sample_iter(&Alphanumeric).take(128).map(char::from).collect();

        Self { keypair: keys.keypair, token: keys.token_address, random }
    }
}

impl Into<Keys> for KeysFile {
    fn into(self) -> Keys {
        Keys { keypair: self.keypair, token_address: self.token }
    }
}

impl KeysFile {
    pub async fn write(self, path: &str) -> Result<(), Box<dyn std::error::Error>> {
        let serialized = serde_json::to_string(&self)?;
        let sealed = detee_sgx::SealingConfig::new()?.seal_data(serialized.into_bytes())?;
        tokio::fs::write(path, sealed).await.map_err(Into::into)
    }

    pub async fn read(path: &str, state: &State) -> Result<Self, Box<dyn std::error::Error>> {
        let sealed = tokio::fs::read(path).await?;
        let serialized = detee_sgx::SealingConfig::new()?.un_seal_data(sealed).map_err(|e| {
            match e {
                detee_sgx::SgxError::UnSealingError(ref ue) => {
                    state.increase_disk_attacks();
                    println!("The disk data is corrupted: {ue}");
                }
                _ => println!("Failed to unseal data: {e}"),
            };
            e
        })?;
        Ok(serde_json::from_str(&String::from_utf8(serialized)?)?)
    }
}

const LOG_PATH: &str = "/host/main/logs";

pub struct Logfile {}

impl Logfile {
    pub fn append(msg: &str) -> Result<(), Box<dyn std::error::Error>> {
        use std::io::Write;
        let mut file = std::fs::OpenOptions::new().create(true).append(true).open(LOG_PATH)?;
        file.write_all(msg.as_bytes())?;
        file.write_all(b"\n")?;
        Ok(())
    }
}
